# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gio

from gettext import gettext as _

from eolie.define import App, LoadingState


class WebViewErrors:
    """
        Implement WebView erros, should be inherited by a WebView
    """

    def __init__(self):
        """
            Init errors
        """
        self.__bad_tls = None  # Keep invalid TLS certificate
        self.connect("load-failed", self.__on_load_failed)
        self.connect("load-failed-with-tls-errors", self.__on_load_failed_tls)
        self.connect("web-process-crashed", self.__on_web_process_crashed)

    def reset_bad_tls(self):
        """
            Reset invalid certificate
        """
        self.__bad_tls = None

    @property
    def bad_tls(self):
        """
            Get invalid certificate
            @return Gio.TlsCertificate
        """
        return self.__bad_tls

#######################
# PROTECTED           #
#######################

#######################
# PRIVATE             #
#######################
    def __check_for_network(self, uri):
        """
            Load uri when network is available
        """
        if Gio.NetworkMonitor.get_default().get_network_available():
            self.load_uri(uri)
        else:
            return True

    def __on_load_failed(self, view, event, uri, error):
        """
            Show error page
            @param view as WebKit2.WebView
            @param event as WebKit2.LoadEvent
            @param uri as str
            @param error as GLib.Error
        """
        self._loading_state = LoadingState.ERROR
        # Ignore HTTP errors
        if error.code > 101:
            return False
        network_available = Gio.NetworkMonitor.get_default(
        ).get_network_available()
        f = Gio.File.new_for_uri("resource:///org/gnome/Eolie/error.css")
        (status, css_content, tag) = f.load_contents(None)
        css = css_content.decode("utf-8")
        # Hide reload button if network is down
        if network_available:
            css = css.replace("@button@", "")
        else:
            css = css.replace("@button@", "display: none")
        f = Gio.File.new_for_uri("resource:///org/gnome/Eolie/error.html")
        (status, content, tag) = f.load_contents(None)
        html = content.decode("utf-8")
        if network_available:
            title = _("Failed to load this web page")
            detail = _("It may be temporarily inaccessible or moved"
                       " to a new address.<br/>"
                       "You may wish to verify that your internet"
                       " connection is working correctly.")
            icon = "dialog-information-symbolic"
        else:
            title = _("Network not available")
            detail = _("Check your network connection")
            icon = "network-offline-symbolic"
        html = html % (title,
                       css,
                       uri,
                       "internal://%s" % icon,
                       title,
                       "<b>%s</b> %s" % (uri, error.message),
                       detail,
                       "suggested-action",
                       _("Retry"))
        self.load_alternate_html(html, uri)
        if not network_available:
            GLib.timeout_add(1000, self.__check_for_network, uri)
        return True

    def __on_load_failed_tls(self, view, uri, certificate, errors):
        """
            Show TLS error page
            @param view as WebKit2.WebView
            @param certificate as Gio.TlsCertificate
            @parma errors as Gio.TlsCertificateFlags
        """
        self._loading_state = LoadingState.ERROR
        self.__bad_tls = certificate
        accept_uri = uri.replace("https://", "accept://")
        if App().websettings.get("accept_tls", uri):
            self.load_uri(accept_uri)
        else:
            f = Gio.File.new_for_uri("resource:///org/gnome/Eolie/error.css")
            (status, css_content, tag) = f.load_contents(None)
            css = css_content.decode("utf-8")
            f = Gio.File.new_for_uri("resource:///org/gnome/Eolie/error.html")
            (status, content, tag) = f.load_contents(None)
            html = content.decode("utf-8")
            if errors == Gio.TlsCertificateFlags.BAD_IDENTITY:
                error = _("The certificate does not match this website")
            elif errors == Gio.TlsCertificateFlags.EXPIRED:
                error = _("The certificate has expired")
            elif errors == Gio.TlsCertificateFlags.UNKNOWN_CA:
                error = _("The signing certificate authority is not known")
            elif errors == Gio.TlsCertificateFlags.GENERIC_ERROR:
                error = _("The certificate contains errors")
            elif errors == Gio.TlsCertificateFlags.REVOKED:
                error = _("The certificate has been revoked")
            elif errors == Gio.TlsCertificateFlags.INSECURE:
                error = _("The certificate is signed using"
                          " a weak signature algorithm")
            elif errors == Gio.TlsCertificateFlags.NOT_ACTIVATED:
                error = _(
                    "The certificate activation time is still in the future")
            else:
                error = _("The identity of this website has not been verified")
            html = html % (_("Connection is not secure"),
                           css,
                           accept_uri,
                           "internal://dialog-warning-symbolic",
                           _("Connection is not secure"),
                           error,
                           _("This does not look like the real %s.<br/>"
                             "Attackers might be trying to steal or alter"
                             " information going to or from this site"
                             " (for example, private messages, credit card"
                             " information, or passwords).") % uri,
                           "destructive-action",
                           _("Accept Risk and Proceed"))
            self.load_alternate_html(html, uri)
        return True

    def __on_web_process_crashed(self, webview):
        """
            We just crashed :-(
            @param webview as WebKit2.WebView
        """
        self._loading_state = LoadingState.ERROR
        f = Gio.File.new_for_uri("resource:///org/gnome/Eolie/error.css")
        (status, css_content, tag) = f.load_contents(None)
        css = css_content.decode("utf-8")
        f = Gio.File.new_for_uri("resource:///org/gnome/Eolie/error.html")
        (status, content, tag) = f.load_contents(None)
        html = content.decode("utf-8")
        html = html % (_("WebKit web engine crashed"),
                       css,
                       "https://bugs.webkit.org/"
                       "enter_bug.cgi?product=WebKit')",
                       "internal://help-faq-symbolic",
                       _("WebKit web engine crashed"),
                       "",
                       _("The webpage was terminated unexpectedly. "
                         "To continue, reload or go to another page.<br/><br/>"
                         "If problem persists, you can report a bug :)<br/>"
                         "Use <b>'WebKit Gtk'</b> as component.<br/>"
                         "Set <b>[GTK]</b> as subject prefix."),
                       "suggested-action",
                       _("Report a bug now"))
        self.load_html(html, webview.uri)
        return True



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# Copyright (c) 2017 Bilal Elmoussaoui <bil.elmoussaoui@gmail.com>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import logging
import sys

from eolie.define import App


class Logger:
    """
        Logger class.
    """
    FORMAT = "[%(levelname)-s] %(asctime)s %(message)s"
    DATE = "%Y-%m-%d %H:%M:%S"
    __log = None
    APP = "org.gnome.Eolie"

    @staticmethod
    def get_default():
        """
            Return default instance of Logger
            @return Logger
        """
        if Logger.__log is None:
            logger = logging.getLogger(Logger.APP)

            handler = logging.StreamHandler(sys.stdout)
            formater = logging.Formatter(Logger.FORMAT, Logger.DATE)
            handler.setFormatter(formater)
            logger.addHandler(handler)
            logger.setLevel(logging.DEBUG)

            Logger.__log = logging.getLogger(Logger.APP)
        return Logger.__log

    @staticmethod
    def warning(msg, *args):
        """
            Log warning message
            @parma msg as str
        """
        Logger.get_default().warning(msg, *args)

    @staticmethod
    def debug(msg, *args):
        """
            Log debug message
            @parma msg as str
        """
        if App().settings.get_value("debug"):
            Logger.get_default().debug(msg, *args)

    @staticmethod
    def sync_debug(msg, *args):
        """
            Log debug sync message
            @parma msg as str
        """
        if App().settings.get_value("debug-sync"):
            Logger.get_default().debug(msg, *args)

    @staticmethod
    def info(msg, *args):
        """
            Log info message
            @parma msg as str
        """
        Logger.get_default().info(msg, *args)

    @staticmethod
    def error(msg, *args):
        """
            Log error message
            @parma msg as str
        """
        Logger.get_default().error(msg, *args)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gdk, GLib, WebKit2

from eolie.sites_manager_child import SitesManagerChild
from eolie.define import App, LoadingType, MARGIN_SMALL
from eolie.utils import get_safe_netloc, update_popover_internals


class SitesManager(Gtk.Grid):
    """
        Site manager (merged netloc of opened pages)
    """

    def __init__(self, window):
        """
            Init stack
            @param window as Window
        """
        Gtk.Grid.__init__(self)
        self.set_orientation(Gtk.Orientation.VERTICAL)
        self.__window = window
        self.__initial_sort = []
        self.__show_labels = False
        self.set_property("width-request", 50)
        self.get_style_context().add_class("sidebar")
        self.__scrolled = Gtk.ScrolledWindow()
        self.__scrolled.set_policy(Gtk.PolicyType.NEVER,
                                   Gtk.PolicyType.AUTOMATIC)
        self.__scrolled.set_vexpand(True)
        self.__scrolled.set_hexpand(True)
        self.__scrolled.show()
        viewport = Gtk.Viewport()
        viewport.show()
        self.__scrolled.add(viewport)
        self.set_hexpand(False)

        self.__box = Gtk.ListBox.new()
        self.__box.set_activate_on_single_click(True)
        self.__box.set_selection_mode(Gtk.SelectionMode.NONE)
        self.__box.set_margin_start(2)
        self.__box.set_margin_end(2)
        self.__box.set_margin_top(2)
        self.__box.set_margin_bottom(2)
        self.__box.show()
        self.__box.connect("row-activated", self.__on_row_activated)

        viewport.set_property("valign", Gtk.Align.START)
        viewport.add(self.__box)

        menu_button = Gtk.Button.new_from_icon_name(
            "view-more-horizontal-symbolic", Gtk.IconSize.BUTTON)
        menu_button.show()
        menu_button.get_style_context().add_class("overlay-button")
        menu_button.set_property("margin", MARGIN_SMALL)
        menu_button.connect("clicked", self.__on_menu_button_clicked)

        self.add(self.__scrolled)
        self.add(menu_button)

    def add_webview(self, webview):
        """
            Add a new web view to monitor
            @param webview as WebView
        """
        # Force update
        if webview.uri:
            self.__on_webview_load_changed(webview,
                                           WebKit2.LoadEvent.COMMITTED)
        webview.connect("load-changed", self.__on_webview_load_changed)
        webview.connect("destroy", self.__on_webview_destroy)

    def remove_webview(self, webview):
        """
            Remove web view from pages manager
            @param webview as WebView
        """
        count = len(self.__box.get_children())
        for site in self.__box.get_children():
            site.remove_webview(webview)
            if site.empty and count > 1:
                site.destroy()
        webview.disconnect_by_func(self.__on_webview_load_changed)
        webview.disconnect_by_func(self.__on_webview_destroy)

    def next(self):
        """
            Show next site
        """
        current = None
        children = self.__box.get_children()
        for child in children:
            if child.get_state_flags() & Gtk.StateFlags.VISITED:
                current = child
            child.unset_state_flags(Gtk.StateFlags.VISITED)
        index = current.get_index()
        if index + 1 < len(children):
            next_row = self.__box.get_row_at_index(index + 1)
        else:
            next_row = self.__box.get_row_at_index(0)
        if next_row is not None:
            next_row.set_state_flags(Gtk.StateFlags.VISITED, False)
            self.__window.container.set_visible_webview(next_row.webviews[0])
            if len(next_row.webviews) == 1:
                self.__window.container.set_expose(False)
            else:
                self.__window.container.pages_manager.set_filter(
                    next_row.netloc)
                self.__window.container.set_expose(True)

    def previous(self):
        """
            Show previous site
        """
        current = None
        children = self.__box.get_children()
        for child in children:
            if child.get_state_flags() & Gtk.StateFlags.VISITED:
                current = child
            child.unset_state_flags(Gtk.StateFlags.VISITED)
        index = current.get_index()
        if index == 0:
            next_row = self.__box.get_row_at_index(len(children) - 1)
        else:
            next_row = self.__box.get_row_at_index(index - 1)
        if next_row is not None:
            next_row.set_state_flags(Gtk.StateFlags.VISITED, False)
            self.__window.container.set_visible_webview(next_row.webviews[0])
            if len(next_row.webviews) == 1:
                self.__window.container.set_expose(False)
            else:
                self.__window.container.pages_manager.set_filter(
                    next_row.netloc)
                self.__window.container.set_expose(True)

    def update_visible_child(self):
        """
            Mark current child as visible
            Unmark all others
        """
        current = self.__window.container.webview
        for child in self.__box.get_children():
            if current in child.webviews:
                child.set_state_flags(Gtk.StateFlags.VISITED, False)
                child.update_favicon()
                # Wait loop empty: will fails otherwise if child just created
                GLib.idle_add(self.__scroll_to_child, child)
            else:
                child.unset_state_flags(Gtk.StateFlags.VISITED)

    def set_initial_sort(self, sort):
        """
            Set initial site sort
            @param sort as [str]
        """
        if sort:
            self.__box.set_sort_func(self.__sort_func)
        else:
            self.__box.set_sort_func(None)
        self.__initial_sort = sort

    def update_shown_state(self, webview):
        """
            Update shown state for webview
            @param webview as WebView
        """
        for child in self.__box.get_children():
            for _webview in child.webviews:
                if _webview == webview:
                    child.indicator_label.mark(webview)
                    return

    def show_labels(self, show):
        """
            Show labels on children
            @param show as bool
        """
        self.__show_labels = show
        for child in self.__box.get_children():
            child.show_label(show)

    @property
    def sort(self):
        """
            Get current sort
            @return [str]
        """
        sort = []
        for child in self.__box.get_children():
            sort.append(child.netloc)
        return sort

#######################
# PRIVATE             #
#######################
    def __sort_func(self, row1, row2):
        """
            Sort rows based on inital sort
            @param row1 as Gtk.ListBoxRow
            @param row2 as Gtk.ListBoxRow
        """
        try:
            index1 = self.__initial_sort.index(row1.netloc)
            index2 = self.__initial_sort.index(row2.netloc)
            return index1 > index2
        except:
            return False

    def __get_index(self, netloc):
        """
            Get child index
            @param netloc as str
            @return int
        """
        # Search current index
        children = self.__box.get_children()
        index = 0
        for child in children:
            if child.netloc == netloc:
                break
            index += 1
        return index

    def __scroll_to_child(self, child):
        """
            Scroll to child
            @param child as SitesManagerChild
        """
        adj = self.__scrolled.get_vadjustment()
        if adj is None:
            return
        value = adj.get_value()
        coordinates = child.translate_coordinates(self.__box, 0, 0)
        if coordinates is None:
            return
        y = coordinates[1]
        if y + child.get_allocated_height() >\
                self.__scrolled.get_allocated_height() + value or\
                y - child.get_allocated_height() < 0 + value:
            self.__scrolled.get_vadjustment().set_value(y)

    def __on_webview_destroy(self, webview):
        """
            Clean children
            @param webview as WebView
        """
        self.remove_webview(webview)

    def __on_webview_load_changed(self, webview, event):
        """
            Update children
            @param webview as WebView
            @param event as WebKit2.LoadEvent
        """
        if event != WebKit2.LoadEvent.COMMITTED:
            return
        netloc = get_safe_netloc(webview.uri)
        child = None
        empty_child = None
        # Do not group by netloc
        if webview.is_ephemeral:
            for site in self.__box.get_children():
                if site.is_ephemeral:
                    child = site
                    break
        else:
            # Search for a child for wanted netloc
            # Clean up any child matching webview, allowing us to reuse it
            for site in self.__box.get_children():
                if site.netloc == netloc and site.is_ephemeral is False:
                    child = site
                else:
                    site.remove_webview(webview)
                    if site.empty:
                        empty_child = site

        if child is None:
            # We need to create a new child
            if empty_child is None:
                child = SitesManagerChild(netloc,
                                          self.__window,
                                          webview.is_ephemeral)
                child.show()
                child.add_webview(webview)
                child.show_label(self.__show_labels)
                self.__box.add(child)
                self.update_visible_child()
            # Use empty child
            else:
                child = empty_child
                child.reset(netloc)
                child.add_webview(webview)
        # We already have a child for this netloc
        else:
            # Webview previous child is empty, destroy it
            if empty_child is not None:
                empty_child.destroy()
            child.add_webview(webview)
            self.update_visible_child()
        # Webview really loaded
        if webview.get_uri() is not None:
            child.on_webview_load_changed(webview, event)

    def __on_row_activated(self, listbox, child):
        """
            Show wanted expose
            @param listbox as Gtk.ListBox
            @param child as SitesManagerChild
        """
        webviews = child.webviews
        if len(webviews) == 1:
            self.__window.container.set_visible_webview(webviews[0])
        else:
            from eolie.pages_manager_list import PagesManagerList
            widget = PagesManagerList(self.__window)
            widget.show()
            widget.populate(webviews)
            popover = Gtk.Popover.new(child)
            popover.set_modal(False)
            self.__window.register(popover)
            popover.get_style_context().add_class("box-shadow")
            popover.set_position(Gtk.PositionType.RIGHT)
            popover.add(widget)
            popover.popup()

    def __on_button_press(self, widget, event):
        """
            Hide popover if visible
            @param widget as Gtk.Widget
            @param event as Gdk.EventButton
        """
        if event.type == Gdk.EventType._2BUTTON_PRESS:
            self.__window.container.add_webview(App().start_page,
                                                LoadingType.FOREGROUND)
        return self.__window.close_popovers()

    def __on_menu_button_clicked(self, button):
        """
            Show pages menu
            @param button as Gtk.Button
        """
        self.__window.close_popovers()
        popover = Gtk.Popover.new_from_model(button, App().pages_menu)
        popover.set_modal(False)
        self.__window.register(popover)
        popover.forall(update_popover_internals)
        popover.popup()



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gdk

from eolie.pages_manager_child import PagesManagerChild
from eolie.utils import get_safe_netloc
from eolie.logger import Logger


class PagesManagerBox(Gtk.EventBox):
    """
        Box linked to a Gtk.Stack
    """

    def __init__(self, window):
        """
            Init stack
            @param window as Window
        """
        Gtk.EventBox.__init__(self)
        self.__window = window
        self.__hovered_child = None
        self.__current_child = None
        # Sort pages by netloc
        self.__sort_pages = []
        self.__allow_left_right_in_entry = False
        self.get_style_context().add_class("sidebar")
        grid = Gtk.Grid()
        grid.set_orientation(Gtk.Orientation.VERTICAL)
        grid.show()
        self.__search_entry = Gtk.SearchEntry.new()
        self.__search_entry.connect("search-changed", self.__on_search_changed)
        self.__search_entry.connect("key-press-event",
                                    self.__on_search_key_press)
        self.__search_entry.connect("button-press-event",
                                    self.__on_search_button_press)
        self.__search_entry.show()
        self.__search_bar = Gtk.SearchBar.new()
        self.__search_bar.add(self.__search_entry)
        self.__scrolled = Gtk.ScrolledWindow()
        self.__scrolled.set_vexpand(True)
        self.__scrolled.set_hexpand(True)
        self.__scrolled.show()
        viewport = Gtk.Viewport()
        viewport.show()
        self.__scrolled.add(viewport)
        self.set_hexpand(False)

        self.__box = Gtk.FlowBox.new()
        self.__box.set_activate_on_single_click(True)
        self.__box.set_selection_mode(Gtk.SelectionMode.NONE)
        self.__box.set_max_children_per_line(1000)
        self.__box.set_filter_func(self.__filter_func)
        self.__box.set_sort_func(self.__sort_func)
        self.__box.show()
        self.__box.connect("child-activated", self._on_child_activated)
        viewport.set_property("valign", Gtk.Align.START)
        viewport.add(self.__box)
        grid.add(self.__scrolled)
        grid.add(self.__search_bar)
        self.add(grid)
        self.__event_controller = Gtk.EventControllerMotion.new(self.__box)
        self.__event_controller.connect("motion", self.__on_box_motion)

    def add_webview(self, webview):
        """
            Add webview to pages manager
            @param view as View
            @return child
        """
        child = PagesManagerChild(webview, self.__window)
        child.show()
        self.__box.add(child)
        return child

    def remove_webview(self, webview):
        """
            Remove view from pages manager
        """
        for child in self.__box.get_children():
            if child.webview == webview:
                child.destroy()
                break

    def update_visible_child(self, visible=None):
        """
            Mark current child or visible if passed as arg
            Unmark all others
            @param visible as WebView
        """
        if visible is None:
            visible = self.__window.container.webview
        for child in self.__box.get_children():
            style_context = child.get_style_context()
            if child.webview == visible:
                child.set_state_flags(Gtk.StateFlags.VISITED, False)
                self.__current_child = child
            else:
                child.unset_state_flags(Gtk.StateFlags.VISITED)
                style_context.remove_class("item-selected")

    def update_shown_state(self, webview):
        """
            Update shown state for webview
            @param webview as WebView
        """
        for child in self.__box.get_children():
            if child.webview == webview:
                child.indicator_label.mark(webview)
                return

    def search_grab_focus(self):
        """
            Grab focus on search entry
        """
        self.__search_entry.grab_focus()

    def update_sort(self, sort=[]):
        """
            Reset sort
            @param group as [str]
        """
        self.__sort_pages = sort
        self.__box.invalidate_sort()

    def set_filter(self, search):
        """
            Filter webview
            @param search as str
        """
        self.__search_entry.set_text(search)
        self.__box.invalidate_filter()

    def set_filtered(self, b):
        """
            Show filtering widget
            @param b as bool
        """
        if b and not self.__search_bar.is_visible():
            self.__search_bar.show()
            self.__search_entry.grab_focus()
        elif self.__search_bar.is_visible():
            self.__search_bar.hide()
        self.__search_bar.set_search_mode(b)

    def next(self, switch=True):
        """
            Show next webview
            @param switch as bool
        """
        children = self.__box.get_children()
        if not children:
            return
        # Search for next child
        count = len(children)
        current_index = self.__current_child.get_index()
        wanted_index = current_index + 1
        if wanted_index >= count:
            wanted_index = 0
        child = self.__box.get_child_at_index(wanted_index)
        if switch:
            self.__window.container.set_visible_webview(child.webview)
        else:
            self.update_visible_child(child.webview)

    def previous(self, switch=True):
        """
            Show previous webview
            @param switch as bool
        """
        children = self.__box.get_children()
        if not children:
            return
        # Search for next child
        count = len(children)
        current_index = self.__current_child.get_index()
        wanted_index = current_index - 1
        if wanted_index < 0:
            wanted_index = count - 1
        child = self.__box.get_child_at_index(wanted_index)
        if switch:
            self.__window.container.set_visible_webview(child.webview)
        else:
            self.update_visible_child(child.webview)

    @property
    def filter(self):
        """
            Get filter
            @return str
        """
        return self.__search_entry.get_text()

    @property
    def filtered(self):
        """
            True if filtered
            @return bool
        """
        return self.__search_bar.get_search_mode()

    @property
    def children(self):
        """
            Get children ordered
            @return [PagesManagerChild]
        """
        return self.__box.get_children()

#######################
# PROTECTED           #
#######################
    def _on_child_activated(self, flowbox, child):
        """
            Show wanted webview
            @param flowbox as Gtk.FlowBox
            @param child as PagesManagerChild
        """
        self.__window.close_popovers()
        self.__window.container.set_visible_webview(child.webview)
        self.__window.container.set_expose(False)

#######################
# PRIVATE             #
#######################
    def __get_column_count(self):
        """
            Calculate column count
            @return int
        """
        count = 0
        current_y = None
        for child in self.__box.get_children():
            (x, y) = child.translate_coordinates(self.__box, 0, 0)
            if current_y is not None and current_y != y:
                break
            current_y = y
            count += 1
        return count

    def __get_index(self, webview):
        """
            Get webview index
            @param webview as WebView
            @return int
        """
        # Search current index
        children = self.__box.get_children()
        index = 0
        for child in children:
            if child.webview == webview:
                break
            index += 1
        return index

    def __sort_func(self, row1, row2):
        """
            Sort listbox
            @param row1 as Row
            @param row2 as Row
        """
        try:
            # Group pages by net location then atime
            if self.__sort_pages and row2.webview.uri is not None and\
                    row1.webview.uri is not None:
                netloc1 = get_safe_netloc(row1.webview.uri)
                netloc2 = get_safe_netloc(row2.webview.uri)
                if netloc1 != netloc2 and netloc1 in self.__sort_pages and\
                        netloc2 in self.__sort_pages:
                    index1 = self.__sort_pages.index(netloc1)
                    index2 = self.__sort_pages.index(netloc2)
                    return index2 < index1
                else:
                    return row2.webview.atime > row1.webview.atime
            # Always show current first
            elif self.__current_child is not None and\
                    self.__current_child in [row1, row2]:
                return self.__current_child == row2
            # Unshown first
            elif not row2.webview.shown and row1.webview.shown:
                return True
            else:
                return row2.webview.atime > row1.webview.atime
        except Exception as e:
            Logger.error("PagesManager::__sort_func(): %s", e)

    def __filter_func(self, row):
        """
            Filter list based on current filter
            @param row as Row
        """
        filter = self.__search_entry.get_text()
        if not filter:
            return True
        uri = row.webview.uri
        title = row.webview.title
        if (uri is not None and uri.find(filter) != -1) or\
                (title is not None and title.find(filter) != -1) or\
                (filter == "private://" and row.webview.ephemeral):
            return True
        return False

    def __unselect_selected(self):
        """
            Unselect selected child
        """
        if self.__hovered_child is not None:
            self.__hovered_child.unset_state_flags(
                Gtk.StateFlags.PRELIGHT)
            self.__hovered_child = None

    def __on_box_motion(self, event_controller, x, y):
        """
            Update current selected child
            @param event_controller as Gtk.EventControllerMotion
            @param x as int
            @param y as int
        """
        child = self.__box.get_child_at_pos(x, y)
        if child == self.__hovered_child:
            return
        elif child is not None:
            child.set_state_flags(Gtk.StateFlags.PRELIGHT, False)
            self.__unselect_selected()
            self.__hovered_child = child
        else:
            self.__unselect_selected()

    def __on_search_changed(self, entry):
        """
            Update filter
            @param entry as Gtk.Entry
        """
        self.__box.invalidate_filter()

    def __on_search_button_press(self, entry, event):
        """
            Allow user to use <- and -> in entry
            @param entry as Gtk.Entry
            @param event as Gdk.EventButton
        """
        self.__allow_left_right_in_entry = True

    def __on_search_key_press(self, entry, event):
        """
            Forward event to self if event is <- or -> and entry is empty
            @param entry as Gtk.Entry
            @param event as Gdk.EventKey
        """
        if event.keyval == Gdk.KEY_Left and\
                not self.__allow_left_right_in_entry:
            self.previous(False)
            return True
        elif event.keyval == Gdk.KEY_Right and\
                not self.__allow_left_right_in_entry:
            self.next(False)
            return True
        elif event.keyval in [Gdk.KEY_Down, Gdk.KEY_Up]:
            if self.__current_child is not None:
                column_count = self.__get_column_count()
                index = self.__current_child.get_index()
                if event.keyval == Gdk.KEY_Down:
                    new_index = index + column_count
                else:
                    new_index = index - column_count
                wanted_child = self.__box.get_child_at_index(new_index)
                if wanted_child is not None:
                    self.update_visible_child(wanted_child.webview)
        elif event.keyval in [Gdk.KEY_Return, Gdk.KEY_KP_Enter]:
            if self.__current_child is not None:
                self._on_child_activated(self.__box, self.__current_child)
        elif event.keyval == Gdk.KEY_Escape:
            self.__search_entry.set_text("")



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk

from gettext import gettext as _

from eolie.define import App, Indicator


class GeolocationPopover(Gtk.Popover):
    """
        Show JavaScript message
        @warning: will block current execution
    """

    def __init__(self, uri, request, window):
        """
            Init popover
            @param uri as str
            @param request as WebKit2.PermissionRequest
            @param window as window
        """
        Gtk.Popover.__init__(self)
        self.set_modal(False)
        window.register(self)
        self.__uri = uri
        self.__request = request
        self.__window = window
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Eolie/PopoverGeolocation.ui")
        builder.connect_signals(self)
        self.__switch = builder.get_object("switch")
        widget = builder.get_object("widget")
        label = builder.get_object("label")
        label.set_text(_("Grant this page access to your location?"))
        self.add(widget)

#######################
# PROTECTED           #
#######################
    def _on_ok_button_clicked(self, button):
        """
            Pass ok to js
            @param button as Gtk.Button
        """
        self.__request.allow()
        self.hide()
        if self.__switch.get_active():
            App().websettings.set("geolocation", self.__uri, True)
            self.__window.toolbar.title.show_indicator(Indicator.GEOLOCATION)

    def _on_cancel_button_clicked(self, button):
        """
            Pass ok to js
            @param button as Gtk.Button
        """
        self.__request.deny()
        self.hide()

#######################
# PRIVATE             #
#######################



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gdk, Gio, WebKit2, GLib

from gettext import gettext as _
from urllib.parse import urlparse
from time import time

from eolie.define import App, LoadingType
from eolie.utils import get_snapshot
from eolie.search import Search
from eolie.logger import Logger


class WebViewMenuSignals:
    """
        Handle webview menu signals
    """

    def __init__(self):
        """
            Init class
        """
        self.connect("context-menu", self.__on_context_menu)
        self.connect("context-menu-dismissed",
                     self.__on_context_menu_dismissed)

#######################
# PRIVATE             #
#######################
    def __on_context_menu(self, webview, context_menu, event, hit):
        """
            Add custom items to menu
            @param webview as WebView
            @param context_menu as WebKit2.ContextMenu
            @param event as Gdk.Event
            @param hit as WebKit2.HitTestResult
        """
        parsed = urlparse(webview.uri)
        if parsed.scheme == "populars":
            context_menu.remove_all()
            if hit.context_is_link():
                action = Gio.SimpleAction(name="reload_preview")
                App().add_action(action)
                action.connect("activate",
                               self.__on_reload_preview_activate,
                               hit.get_link_uri())
                item = WebKit2.ContextMenuItem.new_from_gaction(
                    action,
                    _("Reload preview"),
                    None)
                context_menu.append(item)
                return
        position = 0
        for item in context_menu.get_items():
            action = item.get_stock_action()
            if action in [WebKit2.ContextMenuAction.OPEN_LINK,
                          WebKit2.ContextMenuAction.OPEN_VIDEO_IN_NEW_WINDOW,
                          WebKit2.ContextMenuAction.OPEN_AUDIO_IN_NEW_WINDOW,
                          WebKit2.ContextMenuAction.OPEN_FRAME_IN_NEW_WINDOW]:
                context_menu.remove(item)
            elif action == WebKit2.ContextMenuAction.OPEN_LINK_IN_NEW_WINDOW:
                context_menu.remove(item)
                action = Gio.SimpleAction(name="open_new_page")
                App().add_action(action)
                action.connect("activate",
                               self.__on_open_new_page_activate,
                               hit.get_link_uri(), False)
                item = WebKit2.ContextMenuItem.new_from_gaction(
                    action,
                    _("Open link in a new page"),
                    None)
                context_menu.insert(item, position)
                action = Gio.SimpleAction(name="open_new_private_page")
                App().add_action(action)
                action.connect("activate",
                               self.__on_open_new_page_activate,
                               hit.get_link_uri(), True)
                item = WebKit2.ContextMenuItem.new_from_gaction(
                    action,
                    _("Open link in a new private page"),
                    None)
                context_menu.insert(item, position + 1)
                action = Gio.SimpleAction(name="open_new_window")
                App().add_action(action)
                action.connect("activate",
                               self.__on_open_new_window_activate,
                               hit.get_link_uri())
                item = WebKit2.ContextMenuItem.new_from_gaction(
                    action,
                    _("Open link in a new window"),
                    None)
                context_menu.append(item)
                item = WebKit2.ContextMenuItem.new_separator()
                context_menu.insert(item, position + 2)

        # Get current selection
        selection = ""
        user_data = context_menu.get_user_data()
        if user_data is not None and user_data.get_string():
            selection = user_data.get_string()
        if selection and hit.context_is_selection():
            action = Gio.SimpleAction(name="search_words")
            App().add_action(action)
            action.connect("activate",
                           self.__on_search_words_activate,
                           selection)
            item = WebKit2.ContextMenuItem.new_from_gaction(
                action,
                _("Search on the Web"),
                None)
            context_menu.append(item)
        if not hit.context_is_link() and parsed.scheme in ["http", "https"]:
            item = WebKit2.ContextMenuItem.new_separator()
            context_menu.append(item)
            # Save page as image
            action = Gio.SimpleAction(name="save_as_image")
            App().add_action(action)
            action.connect("activate",
                           self.__on_save_as_image_activate)
            item = WebKit2.ContextMenuItem.new_from_gaction(
                action,
                _("Save page as image"),
                None)
            context_menu.append(item)

    def __on_context_menu_dismissed(self, webview):
        """
            Add custom items to menu
            @param webview as WebView
        """
        self._last_click_time = time()

    def __on_open_new_page_activate(self, action, variant, uri, ephemeral):
        """
            Open link in a new page
            @param action as Gio.SimpleAction
            @param variant as GLib.Variant
            @param uri as str
            @param ephemeral as bool
        """
        self.window.container.add_webview_for_uri(
                                          uri,
                                          LoadingType.FOREGROUND,
                                          self.is_ephemeral or ephemeral)

    def __on_open_new_window_activate(self, action, variant, uri):
        """
            Open link in a new page
            @param action as Gio.SimpleAction
            @param variant as GLib.Variant
            @param uri as str
            @param ephemeral as bool
        """
        window = App().get_new_window()
        window.container.add_webview(uri,
                                     LoadingType.FOREGROUND,
                                     self.ephemeral)

    def __on_search_words_activate(self, action, variant, selection):
        """
            Open link in a new page
            @param action as Gio.SimpleAction
            @param variant as GLib.Variant
            @param selection as str
        """
        search = Search()
        uri = search.get_search_uri(selection)
        self.window.container.add_webview(uri, LoadingType.FOREGROUND)

    def __on_save_as_image_activate(self, action, variant):
        """
            Save image in /tmp and show it to user
            @param action as Gio.SimpleAction
            @param variant as GLib.Variant
        """
        self.get_snapshot(WebKit2.SnapshotRegion.FULL_DOCUMENT,
                          WebKit2.SnapshotOptions.NONE,
                          None,
                          self.__on_snapshot,
                          True)

    def __on_reload_preview_activate(self, action, variant, uri):
        """
            Reload preview for uri
            @param action as Gio.SimpleAction
            @param variant as GLib.Variant
            @param uri as str
        """
        try:
            webview = WebKit2.WebView.new()
            webview.show()
            window = Gtk.OffscreenWindow.new()
            window.set_size_request(self.get_allocated_width(),
                                    self.get_allocated_height())
            window.add(webview)
            window.show()
            webview.load_uri(uri)
            webview.connect("load-changed", self.__on_load_changed, uri)
        except Exception as e:
            Logger.error("""WebViewMenuSignals::
                            __on_reload_preview_activate(): %s""", e)

    def __on_load_changed(self, webview, event, uri):
        """
            Get a snapshot
            @param webview as WebView
            @param event as WebKit2.LoadEvent
            @param uri as str
        """
        if event == WebKit2.LoadEvent.FINISHED:
            GLib.timeout_add(3000,
                             webview.get_snapshot,
                             WebKit2.SnapshotRegion.FULL_DOCUMENT,
                             WebKit2.SnapshotOptions.NONE,
                             None,
                             get_snapshot,
                             self.__on_preview_snapshot,
                             webview,
                             uri, True)

    def __on_preview_snapshot(self, surface, webview, uri, first_pass):
        """
            Cache snapshot
            @param surface as cairo.Surface
            @param webview as WebKit2.WebView
            @param uri as str
            @param first_pass as bool
        """
        # The 32767 limit on the width/height dimensions
        # of an image surface is new in cairo 1.10,
        # try with WebKit2.SnapshotRegion.VISIBLE
        if surface is None and first_pass:
            webview.get_snapshot(WebKit2.SnapshotRegion.VISIBLE,
                                 WebKit2.SnapshotOptions.NONE,
                                 None,
                                 get_snapshot,
                                 self.__on_preview_snapshot,
                                 webview,
                                 uri, False)
        else:
            if surface is not None:
                App().art.save_artwork(uri, surface, "start")
                self.reload()
            window = webview.get_toplevel()
            webview.destroy()
            window.destroy()

    def __on_snapshot(self, webview, result, first_pass):
        """
            Set snapshot on main image
            @param webview as WebView
            @param result as Gio.AsyncResult
            @param first_pass as bool
        """
        try:
            snapshot = webview.get_snapshot_finish(result)
            pixbuf = Gdk.pixbuf_get_from_surface(snapshot, 0, 0,
                                                 snapshot.get_width(),
                                                 snapshot.get_height())
            pixbuf.savev("/tmp/eolie_snapshot.png", "png", [None], [None])
            Gtk.show_uri_on_window(self.window,
                                   "file:///tmp/eolie_snapshot.png",
                                   Gtk.get_current_event_time())
        except Exception as e:
            Logger.error("WebView::__on_snapshot(): %s", e)
            # The 32767 limit on the width/height dimensions
            # of an image surface is new in cairo 1.10,
            # try with WebKit2.SnapshotRegion.VISIBLE
            if first_pass:
                self.get_snapshot(WebKit2.SnapshotRegion.VISIBLE,
                                  WebKit2.SnapshotOptions.NONE,
                                  None,
                                  self.__on_snapshot,
                                  False)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GObject, GLib, Soup

from eolie.css_rule_list import CSSRuleList
from eolie.logger import Logger


class StyleSheet(GObject.Object):
    """
        Represent a stylesheet
    """

    __gsignals__ = {
        "populated": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self, uri=None, contents=None, cancellable=None):
        """
            Init StyleSheet
            @param uri as str
            @param contents as str
            @param cancellable as Gio.Cancellable
        """
        GObject.Object.__init__(self)
        if cancellable is None:
            self.__cancellable = Gio.Cancellable.new()
        else:
            self.__cancellable = cancellable
        self.__uri = uri
        self.__contents = contents
        self.__css_rules = None
        self.__css_text = None
        self.__started_time = 0

    def populate(self):
        """
            Populate styleheet
        """
        if self.__uri is not None and self.__contents is None:
            self.__contents = self.__get_uri_contents(self.__uri)
        if self.__contents is not None:
            self.__css_rules = CSSRuleList(self.__contents,
                                           self.__uri,
                                           self.__cancellable)
        GLib.idle_add(self.emit, "populated")

    def set_css_text(self, css_text):
        """
            Set css text
            @param css_text as str
        """
        self.__css_text = css_text

    def set_started_time(self, started_time):
        """
            Set started time
            @param started time as int
        """
        self.__started_time = started_time

    @property
    def started_time(self):
        """
            Get started time
            @return int
        """
        return self.__started_time

    @property
    def css_text(self):
        """
            Get css text
            @return str
        """
        if self.__css_text is not None:
            return self.__css_text
        elif self.__css_rules is not None:
            self.__css_text = self.__css_rules.css_text
            return self.__css_text
        return ""

    @property
    def uri(self):
        """
            Get stylesheet URI
            @return str
        """
        return self.__uri

    @property
    def contents(self):
        """
            Get contents
            @return str
        """
        return self.__contents

    @property
    def populated(self):
        """
            True if stylesheet is populated
            @return bool
        """
        return self.__css_text is not None or (
            self.__css_rules is not None and self.__css_rules.populated)

#######################
# PRIVATE             #
#######################
    def __get_uri_contents(self, uri):
        """
            Get URI content
            @param uri as str
            @return str
        """
        try:
            session = Soup.Session.new()
            msg = Soup.Message.new("GET", uri)
            content = session.send_and_read(msg, self.__cancellable)
            try:
                return content.decode("utf-8")
            except:
                return content.decode("iso8859-1")
        except Exception as e:
            Logger.error("StyleSheet::__get_uri_contents(): %s -> %s" %
                         (e, uri))
        return None



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import gi
gi.require_version('Secret', '1')
from gi.repository import Secret, GLib

from eolie.logger import Logger


class PasswordsHelper:
    """
        Simpler helper for Secret
    """

    def __init__(self):
        """
            Init helper
        """
        # Initial password lookup, prevent a lock issue in Flatpak backend
        if GLib.file_test("/app", GLib.FileTest.EXISTS):
            SecretSchema = {"type": Secret.SchemaAttributeType.STRING}
            SecretAttributes = {"type": "eolie web login"}
            schema = Secret.Schema.new("org.gnome.Eolie",
                                       Secret.SchemaFlags.NONE,
                                       SecretSchema)
            Secret.password_lookup_sync(schema, SecretAttributes)

    def get_all(self, callback, *args):
        """
            Call function
            @param callback as function
            @param args
        """
        try:
            SecretSchema = {
                "type": Secret.SchemaAttributeType.STRING,
            }
            SecretAttributes = {
                "type": "eolie web login",
            }
            schema = Secret.Schema.new("org.gnome.Eolie",
                                       Secret.SchemaFlags.NONE,
                                       SecretSchema)
            Secret.password_search(schema, SecretAttributes,
                                   Secret.SearchFlags.ALL |
                                   Secret.SearchFlags.UNLOCK |
                                   Secret.SearchFlags.LOAD_SECRETS,
                                   None,
                                   self.__on_secret_search,
                                   None,
                                   callback,
                                   *args)
        except Exception as e:
            Logger.debug("PasswordsHelper::get_all(): %s", e)

    def get(self, uri, userform, passform, callback, *args):
        """
            Call function
            @param uri as str
            @param userform as str
            @param passform as str
            @param callback as function
            @param args
        """
        try:
            SecretSchema = {
                "type": Secret.SchemaAttributeType.STRING,
                "hostname": Secret.SchemaAttributeType.STRING,
                "userform": Secret.SchemaAttributeType.STRING,
            }
            SecretAttributes = {
                "type": "eolie web login",
                "hostname": uri,
                "userform": userform,
            }
            if passform is not None:
                SecretSchema["passform"] = Secret.SchemaAttributeType.STRING
                SecretAttributes["passform"] = passform

            schema = Secret.Schema.new("org.gnome.Eolie",
                                       Secret.SchemaFlags.NONE,
                                       SecretSchema)
            Secret.password_search(schema, SecretAttributes,
                                   Secret.SearchFlags.ALL |
                                   Secret.SearchFlags.UNLOCK |
                                   Secret.SearchFlags.LOAD_SECRETS,
                                   None,
                                   self.__on_secret_search,
                                   uri,
                                   callback,
                                   *args)
        except Exception as e:
            Logger.debug("PasswordsHelper::get(): %s", e)

    def get_by_uuid(self, uuid, callback, *args):
        """
            Get password attributes by uri
            @param uuid as str
            @param callback as function
            @param args
        """
        try:
            SecretSchema = {
                "type": Secret.SchemaAttributeType.STRING,
                "uuid": Secret.SchemaAttributeType.STRING,
            }
            SecretAttributes = {
                "type": "eolie web login",
                "uuid": uuid,
            }

            schema = Secret.Schema.new("org.gnome.Eolie",
                                       Secret.SchemaFlags.NONE,
                                       SecretSchema)
            Secret.password_search(schema, SecretAttributes,
                                   Secret.SearchFlags.ALL |
                                   Secret.SearchFlags.UNLOCK |
                                   Secret.SearchFlags.LOAD_SECRETS,
                                   None,
                                   self.__on_secret_search,
                                   uuid,
                                   callback,
                                   *args)
        except Exception as e:
            Logger.debug("PasswordsHelper::get_by_uuid(): %s", e)

    def get_sync(self, callback, *args):
        """
            Get sync password
            @param callback as function
        """
        try:
            SecretSchema = {
                "sync": Secret.SchemaAttributeType.STRING
            }
            SecretAttributes = {
                "sync": "mozilla"
            }
            schema = Secret.Schema.new("org.gnome.Eolie",
                                       Secret.SchemaFlags.NONE,
                                       SecretSchema)
            Secret.password_search(schema, SecretAttributes,
                                   Secret.SearchFlags.UNLOCK |
                                   Secret.SearchFlags.LOAD_SECRETS,
                                   None,
                                   self.__on_secret_search,
                                   None,
                                   callback,
                                   *args)
        except Exception as e:
            Logger.debug("PasswordsHelper::get_sync(): %s", e)

    def store(self, user_form_name, user_form_value, pass_form_name,
              pass_form_value, hostname_uri, uri, uuid, callback, *args):
        """
            Store password
            @param user_form_name as str
            @param user_form_value as str
            @param pass_form_name as str
            @param pass_form_value as str
            @param hostname_uri as str
            @param uri as str
            @param uuid as str
            @param callback as function
        """
        # seems to happen, thanks firefox
        if uri is None:
            return
        try:
            # Clear item if exists
            SecretSchema = {
                "type": Secret.SchemaAttributeType.STRING,
                "login": Secret.SchemaAttributeType.STRING,
                "hostname": Secret.SchemaAttributeType.STRING,
                "formSubmitURL": Secret.SchemaAttributeType.STRING,
                "userform": Secret.SchemaAttributeType.STRING,
                "passform": Secret.SchemaAttributeType.STRING,
            }
            SecretAttributes = {
                "type": "eolie web login",
                "login": user_form_value,
                "hostname": hostname_uri,
                "formSubmitURL": uri,
                "userform": user_form_name,
                "passform": pass_form_name
            }
            schema = Secret.Schema.new("org.gnome.Eolie",
                                       Secret.SchemaFlags.NONE,
                                       SecretSchema)
            Secret.password_clear(schema,
                                  SecretAttributes,
                                  None,
                                  self.__on_clear_search,
                                  callback,
                                  *args)

            schema_string = "org.gnome.Eolie: %s > %s" % (user_form_value,
                                                          hostname_uri)
            SecretSchema = {
                "type": Secret.SchemaAttributeType.STRING,
                "uuid": Secret.SchemaAttributeType.STRING,
                "login": Secret.SchemaAttributeType.STRING,
                "hostname": Secret.SchemaAttributeType.STRING,
                "formSubmitURL": Secret.SchemaAttributeType.STRING,
                "userform": Secret.SchemaAttributeType.STRING,
                "passform": Secret.SchemaAttributeType.STRING,
            }
            SecretAttributes = {
                "type": "eolie web login",
                "uuid": uuid,
                "login": user_form_value,
                "hostname": hostname_uri,
                "formSubmitURL": uri,
                "userform": user_form_name,
                "passform": pass_form_name
            }

            schema = Secret.Schema.new("org.gnome.Eolie",
                                       Secret.SchemaFlags.NONE,
                                       SecretSchema)
            Secret.password_store(schema, SecretAttributes,
                                  Secret.COLLECTION_DEFAULT,
                                  schema_string,
                                  pass_form_value,
                                  None,
                                  callback)
        except Exception as e:
            Logger.debug("PasswordsHelper::store(): %s", e)

    def store_sync(self, login, password, callback=None, *args):
        """
            Store Firefox Sync password
            @param login as str
            @param password as str
            @param callback as function
            @param data
        """
        try:
            schema_string = "org.gnome.Eolie.sync"
            SecretSchema = {
                "sync": Secret.SchemaAttributeType.STRING,
                "login": Secret.SchemaAttributeType.STRING,
            }
            schema = Secret.Schema.new("org.gnome.Eolie",
                                       Secret.SchemaFlags.NONE,
                                       SecretSchema)
            SecretAttributes = {
                "sync": "mozilla",
                "login": login,
            }
            Secret.password_store(schema, SecretAttributes,
                                  Secret.COLLECTION_DEFAULT,
                                  schema_string,
                                  password,
                                  None,
                                  callback,
                                  *args)
        except Exception as e:
            Logger.debug("PasswordsHelper::store_sync(): %s", e)

    def clear(self, uuid, callback=None, *args):
        """
            Clear password
            @param uuid as str
            @param callback as function
        """
        try:
            SecretSchema = {
                "type": Secret.SchemaAttributeType.STRING,
                "uuid": Secret.SchemaAttributeType.STRING
            }
            SecretAttributes = {
                "type": "eolie web login",
                "uuid": uuid
            }
            schema = Secret.Schema.new("org.gnome.Eolie",
                                       Secret.SchemaFlags.NONE,
                                       SecretSchema)
            Secret.password_clear(schema,
                                  SecretAttributes,
                                  None,
                                  self.__on_clear_search,
                                  callback,
                                  *args)
        except Exception as e:
            Logger.debug("PasswordsHelper::clear(): %s", e)

    def clear_sync(self, callback, *args):
        """
            Clear sync secrets
            @param callback as function
        """
        try:
            SecretSchema = {
                "sync": Secret.SchemaAttributeType.STRING
            }
            SecretAttributes = {
                "sync": "mozilla"
            }
            schema = Secret.Schema.new("org.gnome.Eolie",
                                       Secret.SchemaFlags.NONE,
                                       SecretSchema)
            Secret.password_clear(schema,
                                  SecretAttributes,
                                  None,
                                  self.__on_clear_search,
                                  callback,
                                  *args)
        except Exception as e:
            Logger.debug("PasswordsHelper::clear_sync(): %s", e)

    def clear_all(self):
        """
            Clear passwords
        """
        try:
            SecretSchema = {
                "type": Secret.SchemaAttributeType.STRING
            }
            SecretAttributes = {
                "type": "eolie web login"
            }
            schema = Secret.Schema.new("org.gnome.Eolie",
                                       Secret.SchemaFlags.NONE,
                                       SecretSchema)
            Secret.password_clear(schema,
                                  SecretAttributes,
                                  None,
                                  self.__on_clear_search)
        except Exception as e:
            Logger.debug("PasswordsHelper::clear_all(): %s", e)

#######################
# PRIVATE             #
#######################
    def __on_clear_search(self, source, result, callback=None, *args):
        """
            Clear passwords
            @param source as GObject.Object
            @param result as Gio.AsyncResult
            @param callback as function
        """
        try:
            if result is not None:
                Secret.password_clear_finish(result)
            if callback is not None:
                callback(*args)
        except Exception as e:
            Logger.error("PasswordsHelper::__on_clear_search(): %s" % e)

    def __on_secret_search(self, source, result, uri, callback, *args):
        """
            Set username/password input
            @param source as GObject.Object
            @param result as Gio.AsyncResult
            @param uri as str
            @param callback as function
            @param args
        """
        try:
            items = []
            if result is not None:
                items = Secret.password_search_finish(result)
                count = len(items)
                index = 0
                for item in items:
                    attributes = item.get_attributes()
                    secret = item.retrieve_secret_sync()
                    callback(attributes,
                             secret.get().decode('utf-8'),
                             uri,
                             index,
                             count,
                             *args)
                    index += 1
            if not items:
                callback(None, None, uri, 0, 0, *args)
        except Exception as e:
            Logger.debug("PasswordsHelper::__on_secret_search(): %s", e)
            callback(None, None, uri, 0, 0, *args)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, WebKit2, GObject, Pango

from eolie.define import MARGIN_SMALL
from eolie.utils import emit_signal
from eolie.logger import Logger


class Row(Gtk.ListBoxRow):
    """
        A row with a title and a checkbox
    """

    def __init__(self, item):
        """
            Init Row
            @param item as WebKit2.WebsiteData
        """
        Gtk.ListBoxRow.__init__(self)
        self.__item = item
        label = Gtk.Label.new(item.get_name())
        label.show()
        label.set_halign(Gtk.Align.START)
        label.set_property("margin", MARGIN_SMALL)
        label.set_ellipsize(Pango.EllipsizeMode.END)
        self.add(label)

    @property
    def item(self):
        """
            Get title
            @return WebKit2.WebsiteData
        """
        return self.__item


class ClearDataDialog(Gtk.Bin):
    """
        A clear data dialog
        THANKS TO EPIPHANY DEVS FOR UI FILE!
    """

    __gsignals__ = {
        "destroy-me": (GObject.SignalFlags.RUN_FIRST, None, ())
    }

    def __init__(self):
        """
            Init widget
        """
        Gtk.Bin.__init__(self)
        self.__search = ""
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Eolie/DialogClearData.ui")
        builder.connect_signals(self)
        self.__search_bar = builder.get_object("search_bar")
        self.__remove_button = builder.get_object("remove_button")
        self.__listbox = builder.get_object("listbox")
        self.__listbox.set_sort_func(self.__sort_func)
        self.__listbox.set_filter_func(self.__filter_func)
        self.__populate()
        self.add(builder.get_object("widget"))

#######################
# PROTECTED           #
#######################
    def _on_back_clicked(self, button):
        """
            Ask to be destroyed
            @param button as Gtk.Button
        """
        emit_signal(self, "destroy-me")

    def _on_search_changed(self, entry):
        """
            Filter model
            @param entry as Gtk.SearchEntry
        """
        self.__search = entry.get_text()
        self.__listbox.invalidate_filter()

    def _on_remove_clicked(self, button):
        """
            Clear data
            @param button as Gtk.Button
        """
        context = WebKit2.WebContext.get_default()
        data_manager = context.get_property("website-data-manager")
        rows = self.__listbox.get_selected_rows()
        items = [row.item for row in rows]
        data_manager.remove(WebKit2.WebsiteDataTypes.ALL, items, None)
        for row in rows:
            row.destroy()

    def _on_search_toggled(self, button):
        """
            Show entry
            @param button as Gtk.Button
        """
        self.__search_bar.set_search_mode(button.get_active())

    def _on_row_selected(self, listbox, row):
        """
            Update clear button state
            @param listbox as Gtk.ListBox
            @param row as Gtk.ListBoxRow
        """
        self.__remove_button.set_sensitive(
            len(listbox.get_selected_rows()) != 0)

#######################
# PRIVATE             #
#######################
    def __sort_func(self, rowa, rowb):
        """
            Sort listbox
            @param rowa as Row
            @param rowb as Row
            @return bool
        """
        return rowa.item.get_name() > rowb.item.get_name()

    def __filter_func(self, row):
        """
            Filter listbox
            @param row as Row
            @return bool
        """
        return row.item.get_name().find(self.__search) != -1

    def __populate(self):
        """
            Populate treeview
        """
        context = WebKit2.WebContext.get_default()
        data_manager = context.get_property("website-data-manager")
        data_manager.fetch(WebKit2.WebsiteDataTypes.ALL,
                           None,
                           self.__on_data_manager_fetch)

    def __add_items(self, items):
        """
            Add items to model
            @param items as [WebKit2.WebsiteData]
        """
        if items:
            item = items.pop(0)
            row = Row(item)
            row.show()
            self.__listbox.add(row)
            GLib.idle_add(self.__add_items, items)

    def __on_data_manager_fetch(self, data_manager, result):
        """
            Get fetch result
            @param data_manager as WebKit2.WebsiteDataManager
            @param result as Gio.AsyncResult
        """
        try:
            items = data_manager.fetch_finish(result)
            self.__add_items(items)
        except Exception as e:
            Logger.error("ClearDataDialog::__on_data_manager_fetch(): %s", e)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio

import json

from eolie.define import EOLIE_DATA_PATH
from eolie.logger import Logger


class ContentBlockerExceptions:
    """
        Exception handler
    """
    __JSON_PATH = "%s/content_blocker_json" % EOLIE_DATA_PATH

    def __init__(self, name):
        """
            Init constructor
        """
        try:
            self.__name = name
            self.__rules = []
            f = Gio.File.new_for_path(
                "%s/exceptions_%s.json" % (self.__JSON_PATH, self.__name))
            if f.query_exists():
                (status, contents, tag) = f.load_contents(None)
                if status:
                    self.__rules = json.loads(contents.decode("utf-8"))
        except Exception as e:
            Logger.error("AdblockExceptions::__init__(): %s", e)

    def save(self):
        """
            Save rules to disk
        """
        try:
            f = Gio.File.new_for_path(
                "%s/exceptions_%s.json" % (self.__JSON_PATH, self.__name))
            content = json.dumps(self.__rules)
            f.replace_contents(content.encode("utf-8"),
                               None,
                               False,
                               Gio.FileCreateFlags.REPLACE_DESTINATION,
                               None)
        except Exception as e:
            Logger.error("AdblockExceptions::save(): %s", e)

    def add_domain_exception(self, domain, url_filter=".*", internal=False):
        """
            Add an exception for domain
            @param domain as str
            @param url_filter as str
            @param internal as bool
        """
        if internal:
            rule = self.__get_rule_for_internal_domain(domain, url_filter)
            if rule in self.__rules:
                self.__rules.remove(rule)
        else:
            rule = self.__get_rule_for_domain(domain, url_filter)
            self.__rules.append(rule)

    def remove_domain_exception(self, domain, url_filter=".*", internal=False):
        """
            Remove an exception for domain
            @param domain as str
            @param url_filter as str
            @param internal as bool
        """
        if internal:
            rule = self.__get_rule_for_internal_domain(domain, url_filter)
            self.__rules.append(rule)
        else:
            rule = self.__get_rule_for_domain(domain, url_filter)
            if rule in self.__rules:
                self.__rules.remove(rule)

    def remove_all_domain_exceptions(self, domain):
        """
            Remove all exceptions for a domain
            @param domain as str
        """
        for rule in list(self.__rules):
            if rule["trigger"]["if-domain"] == ["*%s" % domain]:
                self.remove_domain_exception(
                    domain, rule["trigger"]["url-filter"])

    def is_domain_exception(self, domain, url_filter=".*", internal=False):
        """
            True if domain exception exists
            @parma domain as str
            @param url_filter as str
            @param internal as bool
            @return bool
        """
        if internal:
            rule = self.__get_rule_for_internal_domain(domain, url_filter)
            return rule not in self.__rules
        else:
            rule = self.__get_rule_for_domain(domain, url_filter)
            return rule in self.__rules

    @property
    def rules(self):
        """
            Get rules
            @return []
        """
        return self.__rules

#######################
# PRIVATE             #
#######################
    def __get_rule_for_domain(self, domain, url_filter):
        """
            Return rule for domain
            @param domain as str
            @param url_filter as str
            @return {}
        """
        value = "*%s" % domain
        return {
            "trigger": {
                "url-filter": url_filter,
                "if-domain": [value]
            },
            "action": {
                "type": "ignore-previous-rules"
            }
        }

    def __get_rule_for_internal_domain(self, domain, url_filter):
        """
            Return rule for domain
            @param domain as str
            @param url_filter as str
            @return {}
        """
        value = "*%s" % domain
        return {
            "trigger": {
                "url-filter": url_filter,
                "if-domain": [value]
            },
            "action": {
                "type": "block"
            }
        }



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib

from eolie.pages_manager_box import PagesManagerBox


class ExposeContainer:
    """
        Expose management for container
    """

    def __init__(self):
        """
            Init container
        """
        self.__next_timeout_id = None
        self.__previous_timeout_id = None
        self.__expose_stack = Gtk.Stack()
        self.__expose_stack.set_hexpand(True)
        self.__expose_stack.set_vexpand(True)
        self.__expose_stack.set_transition_type(
            Gtk.StackTransitionType.CROSSFADE)
        self.__expose_stack.set_transition_duration(150)
        self.__expose_stack.show()
        self.__pages_manager = PagesManagerBox(self._window)
        self.__pages_manager.show()
        self.overlay.add(self.__expose_stack)
        self.__expose_stack.add_named(self._stack, "stack")
        self.__expose_stack.add_named(self.__pages_manager, "expose")

    def set_expose(self, expose):
        """
            Show current views
            @param expose as bool
        """
        if expose:
            self.__pages_manager.update_sort(self.sites_manager.sort)
            self.__pages_manager.set_filtered(True)
        else:
            self._window.toolbar.actions.view_button.set_active(False)
            self._window.container.pages_manager.set_filter("")
            self.__pages_manager.set_filtered(False)
        self.__set_expose(expose)

    def next(self):
        """
            Show next view
        """
        if self.__next_timeout_id is None and\
                self.__next_timeout_id != -1:
            self.__next_timeout_id = GLib.timeout_add(
                100,
                self.__on_prev_next_timeout,
                self.__pages_manager.next)
        else:
            self.__pages_manager.next()

    def previous(self):
        """
            Show next view
        """
        if self.__previous_timeout_id is None and\
                self.__previous_timeout_id != -1:
            self.__previous_timeout_id = GLib.timeout_add(
                100,
                self.__on_prev_next_timeout,
                self.__pages_manager.previous)
        else:
            self.__pages_manager.previous()

    def ctrl_released(self):
        """
            Disable any pending expose
        """
        if self.__next_timeout_id is not None:
            if self.__next_timeout_id != -1:
                self.pages_manager.next()
                GLib.source_remove(self.__next_timeout_id)
        if self.__previous_timeout_id is not None:
            if self.__previous_timeout_id != -1:
                self.pages_manager.previous()
                GLib.source_remove(self.__previous_timeout_id)

        self.__next_timeout_id = None
        self.__previous_timeout_id = None
        self.set_expose(False)

    @property
    def in_expose(self):
        """
            True if in expose mode
            @return bool
        """
        return self.__expose_stack.get_visible_child_name() == "expose"

    @property
    def pages_manager(self):
        """
            Get pages manager
            @return PagesManager
        """
        return self.__pages_manager

    @property
    def webviews(self):
        """
            Get webviews
            @return webviews as [WebView]
        """
        return self._stack.get_children()

#######################
# PRIVATE             #
#######################
    def __on_prev_next_timeout(self, callback):
        """
            Set expose on and call callback
            @param callback as __next()/__previous()
        """
        self.__next_timeout_id = -1
        self.__previous_timeout_id = -1
        if not self.in_expose:
            self.__set_expose(True)
        callback()

    def __set_expose(self, expose):
        """
            Show current views
            @param expose as bool
            @param search as bool
        """
        # Show expose mode
        if expose:
            self.__expose_stack.set_visible_child_name("expose")
        else:
            self.__expose_stack.set_visible_child_name("stack")
            self.__pages_manager.update_visible_child()
